from __future__ import absolute_import
from builtins import zip
import os
from sqlalchemy.sql import text
from flask import current_app, g
from chill.app import db
from .cache import cache

CHILL_DROP_TABLE_FILES = (
    "drop_chill.sql",
    "drop_query.sql",
    "drop_template.sql",
    "drop_node.sql",
    "drop_node_node.sql",
    "drop_route.sql",
)
CHILL_CREATE_TABLE_FILES = (
    "create_chill.sql",
    "set_current_chill_version.sql",
    "create_query.sql",
    "create_template.sql",
    "create_node.sql",
    "create_node_node.sql",
    "create_route.sql",
)


def init_db():
    """Initialize a new database with the default tables for chill.
    Creates the following tables:
    Chill
    Node
    Node_Node
    Route
    Query
    Template

    The current Chill migration version is added to the Chill table.
    """
    hold_database_readonly_setting = current_app.config.get("database_readonly")
    current_app.config["database_readonly"] = False
    with current_app.app_context():
        for filename in CHILL_CREATE_TABLE_FILES:
            db.execute(text(fetch_query_string(filename)))
    current_app.config["database_readonly"] = hold_database_readonly_setting


def rowify(l, description):
    d = []
    col_names = []
    if l != None and description != None:
        col_names = [x[0] for x in description]
        for row in l:
            d.append(dict(list(zip(col_names, row))))
    return (d, col_names)


def _fetch_sql_string(file_name):
    with current_app.open_resource(os.path.join("queries", file_name), mode="r") as f:
        return f.read()


def fetch_query_string(file_name):
    content = current_app.queries.get(file_name, None)
    if content != None:
        return content
    current_app.logger.info(
        "queries file: '%s' not available. Checking file system..." % file_name
    )

    # folder = current_app.config.get('THEME_SQL_FOLDER', '')
    # file_path = os.path.join(os.path.abspath('.'), folder, file_name)
    folder = current_app.config.get("THEME_SQL_FOLDER")
    file_path = os.path.join(folder, file_name)
    if os.path.isfile(file_path):
        with open(file_path, "r") as f:
            return f.read()
    else:
        # fallback on one that's in app resources
        return _fetch_sql_string(file_name)


def insert_node(**kw):
    "Insert a node with a name and optional value. Return the node id."
    with current_app.app_context():
        result = db.execute(text(fetch_query_string("insert_node.sql")), **kw)
        # TODO: support for postgres may require using a RETURNING id; sql
        # statement and using the inserted_primary_key?
        # node_id = result.inserted_primary_key
        node_id = result.lastrowid
        if not node_id:
            result = result.fetchall()
            node_id = result[0]["id"]
        return node_id


def insert_node_node(**kw):
    """
    Link a node to another node. node_id -> target_node_id.  Where `node_id` is
    the parent and `target_node_id` is the child.
    """
    with current_app.app_context():
        insert_query(name="select_link_node_from_node.sql", node_id=kw.get("node_id"))
        db.execute(text(fetch_query_string("insert_node_node.sql")), **kw)


def delete_node(**kw):
    """
    Delete a node by id.
    """
    with current_app.app_context():
        db.execute(text(fetch_query_string("delete_node_for_id.sql")), **kw)


def select_node(**kw):
    """
    Select node by id.
    """
    with current_app.app_context():
        result = db.execute(
            text(fetch_query_string("select_node_from_id.sql")), **kw
        ).fetchall()
        return result


def insert_route(**kw):
    """
    `path` - '/', '/some/other/path/', '/test/<int:index>/'
    `node_id`
    `weight` - How this path is selected before other similar paths
    `method` - 'GET' is default.
    """
    binding = {"path": None, "node_id": None, "weight": None, "method": "GET"}
    binding.update(kw)
    with current_app.app_context():
        db.execute(text(fetch_query_string("insert_route.sql")), **binding)


def add_template_for_node(name, node_id):
    "Set the template to use to display the node"
    with current_app.app_context():
        db.execute(
            text(fetch_query_string("insert_template.sql")), name=name, node_id=node_id
        )
        result = db.execute(
            text(fetch_query_string("select_template.sql")), name=name, node_id=node_id
        ).fetchall()
        if result:
            template_id = result[0]["id"]
            db.execute(
                text(fetch_query_string("update_template_node.sql")),
                template=template_id,
                node_id=node_id,
            )


def insert_query(**kw):
    """
    Insert a query name for a node_id.
    `name`
    `node_id`

    Adds the name to the Query table if not already there. Sets the query field
    in Node table.
    """
    with current_app.app_context():
        result = db.execute(
            text(fetch_query_string("select_query_where_name.sql")), **kw
        ).fetchall()
        if result:
            kw["query_id"] = result[0]["id"]
        else:
            result = db.execute(text(fetch_query_string("insert_query.sql")), **kw)
            kw["query_id"] = result.lastrowid
            if not kw["query_id"]:
                result = result.fetchall()
                kw["query_id"] = result[0]["id"]
        db.execute(text(fetch_query_string("insert_query_node.sql")), **kw)



import os
import sqlite3

from sqlalchemy.sql import text
from flask import current_app
import yaml

from chill.app import make_app, db
from chill.database import (
        init_db,
        insert_node,
        insert_node_node,
        delete_node,
        select_node,
        insert_route,
        insert_query,
        add_template_for_node,
        fetch_query_string,
        rowify,
        )

def _is_sql_file(file_name):
    folder = current_app.config.get('THEME_SQL_FOLDER')
    file_path = os.path.join(folder, file_name)
    return os.path.isfile(file_path)

def _is_template_file(file_name):
    folder = current_app.config.get('THEME_TEMPLATE_FOLDER')
    file_path = os.path.join(folder, file_name)
    return os.path.isfile(file_path)

def _value_is_simple_string(value):
    if isinstance(value, str):
        _value = value.strip()
        if _value.endswith('.sql') and _is_sql_file(_value):
            return False
        else:
            return True

def _add_node_to_parent(parent_node_id, name, value):
    node_has_template = False
    if isinstance(value, bool):
        raise TypeError('Boolean values are not supported. Surround the value with quotes to set as string.')
    _value = value
    if isinstance(value, int) or isinstance(value, float):
        _value = str(value)
    elif isinstance(value, dict) and set(value.keys()).issubset({'chill_template', 'chill_value'}):
        if value.get('chill_template') == None:
            _add_node_to_parent(parent_node_id, name, value.get('chill_value'))
            return
        else:
            _value = value.get('chill_value')
            if isinstance(_value, int) or isinstance(_value, float):
                _value = str(_value)
            node_has_template = True

    item_node_id = None
    if _value_is_simple_string(_value):
        item_node_id = insert_node(name=name, value=_value)
        insert_query(name='select_link_node_from_node.sql', node_id=item_node_id)
        insert_node_node(node_id=parent_node_id, target_node_id=item_node_id)

    elif isinstance(_value, str) and _is_sql_file(_value):
        item_node_id = insert_node(name=name, value=None)
        insert_query(name=_value, node_id=item_node_id)
        insert_node_node(node_id=parent_node_id, target_node_id=item_node_id)

    else:
        item_node_id = insert_node(name=name, value=None)
        insert_query(name='select_link_node_from_node.sql', node_id=item_node_id)
        insert_node_node(node_id=parent_node_id, target_node_id=item_node_id)

        if isinstance(_value, dict):
            if set(_value.keys()).issubset({'chill_template', 'chill_value'}):
                raise TypeError("chill_template or chill_value should not be set directly under a chill_value")
            for item_name in _value.keys():
                item_value = _value.get(item_name)
                _add_node_to_parent(item_node_id, item_name, item_value)

        elif isinstance(_value, list):
            for item_value in _value:
                _add_node_to_parent(item_node_id, name, item_value)
        elif _value == None:
            pass
        else:
            raise TypeError('unsupported value type. Use only dict, or list.')

    if node_has_template:
        add_template_for_node(value.get('chill_template'), item_node_id)

def _render_chill_node_value(node_id, root=False):
    # get the node
    query_result = db.execute(text(fetch_query_string('select_query_from_node.sql')), {"node_id": node_id}).fetchall()
    value = None
    if query_result:
        #current_app.logger.debug('render {node_id}'.format(node_id=node_id))
        #current_app.logger.debug(query_result)
        values = []
        if len(query_result) > 1:
            raise NotImplementedError('TODO: Support for multiple queries found for a node.')
        for query_name in [x['name'] for x in query_result]:
            if query_name == 'select_link_node_from_node.sql':
                link_nodes_result = db.execute(text(fetch_query_string('select_link_node_from_node.sql')), {"node_id": node_id}).fetchall()
                if link_nodes_result:
                    value = {}
                    if len(link_nodes_result) > 1 and link_nodes_result[0].name == link_nodes_result[1].name:
                        value = []
                        for link_node in link_nodes_result:
                            item = {}
                            item[link_node.name] = _render_chill_node_value(link_node.node_id)
                            value.append(item)
                    else:
                        for link_node in link_nodes_result:
                            value[link_node.name] = _render_chill_node_value(link_node.node_id)

                else:
                    node = select_node(node_id=node_id)[0]
                    value = node.value
            else:
                value = query_name
    else:
        node = select_node(node_id=node_id)[0]
        value = node.value

    if not root:
        template_for_node_result = db.execute(text(fetch_query_string('select_template_from_node.sql')), {"node_id": node_id}).fetchone()
        if template_for_node_result:
            template_for_node = template_for_node_result.name
            chill_value = value
            value = {
                'chill_template': template_for_node,
            }
            if chill_value != None:
                value['chill_value'] = chill_value

    return value

class ChillNode(yaml.YAMLObject):
    yaml_loader = yaml.SafeLoader
    yaml_tag = u'!ChillNode'

    # Default props
    name = None
    value = None
    template = None
    route = None

    def __init__(self, name, node_id=None, value=None, template=None, query=None, path=None, method=None, weight=None):
        self.name = name
        if value != None:
            self.value = value
        if template != None:
            self.template = template
        if query != None:
            if query == 'select_link_node_from_node.sql':
                self.value = _render_chill_node_value(node_id, root=True)
            else:
                self.value = query

        if path != None:
            if isinstance(path, str) and method in (None, 'GET') and weight in (None, ''):
                self.route = path
            else:
                route = {
                    "path": path
                }
                if method != None:
                    route["method"] = method
                if weight != None and weight != '':
                    route["weight"] = weight
                self.route = route

    def load(self):
        if not self.name:
            raise TypeError('the `name` property is required for ChillNode')

        if isinstance(self.value, bool):
            raise TypeError('Boolean values are not supported. Surround the value with quotes to set as string.')

        value = None
        if _value_is_simple_string(self.value):
            value = self.value
        elif isinstance(self.value, int) or isinstance(self.value, float):
            value = str(self.value)

        # Insert the chill_node
        chill_node = insert_node(name=self.name, value=value)

        # Set the route
        if self.route:
            if isinstance(self.route, str):
                insert_route(path=self.route, node_id=chill_node, weight=None, method='GET')
            elif isinstance(self.route, dict):
                route_path = self.route.get('path', None)
                if route_path == None:
                    raise TypeError('A path must be set for a route')
                elif not isinstance(route_path, str):
                    raise TypeError('A path must be a string value')
                route_weight = self.route.get('weight', None)
                route_method = self.route.get('method', 'GET')
                if route_weight != None and not isinstance(route_weight, int):
                    raise TypeError("route weight value needs to be integer if defined")
                if (isinstance(route_method, str) and route_method.upper() not in ('GET', 'POST', 'PUT', 'DELETE', 'PATCH')):
                    raise TypeError("route method value needs to be 'GET', 'POST', 'PUT', 'DELETE', or 'PATCH' if defined")
                elif not isinstance(route_method, str):
                    raise TypeError("route method value needs to be 'GET', 'POST', 'PUT', 'DELETE', or 'PATCH' if defined")
                insert_route(path=self.route['path'], node_id=chill_node, weight=route_weight, method=route_method.upper())

            else:
                raise TypeError('route value other then str or dict not supported')

        # Set the template
        if self.template:
            if isinstance(self.template, str):
                if _is_template_file(self.template):
                    add_template_for_node(name=self.template, node_id=chill_node)
                else:
                    raise TypeError('template value must be a path to a file in template folder')
            else:
                raise TypeError('template value must be a string')

        # Set the query if value is not simple string
        if not value:
            if isinstance(self.value, str) and _is_sql_file(self.value):
                insert_query(name=self.value, node_id=chill_node)

            elif isinstance(self.value, dict):
                if set(self.value.keys()).issubset({'chill_template', 'chill_value'}):
                    raise TypeError("chill_template or chill_value should not be set at the top ChillNode value.")
                for item_name in self.value.keys():
                    item_value = self.value.get(item_name)
                    _add_node_to_parent(chill_node, item_name, item_value)

            elif isinstance(self.value, list):
                for item_value in self.value:
                    _add_node_to_parent(chill_node, 'value', item_value)
            elif self.value == None:
                pass
            else:
                raise TypeError('unsupported value type. Use only dict, or list.')

    def __repr__(self):
        return "%s(name=%r, value=%r, template=%r, route=%r)" % (self.__class__.__name__, self.name, self.value, self.template, self.route)


def dump_yaml(yaml_file):
    "Dump chill database structure to ChillNode yaml objects."

    result = db.execute(text(fetch_query_string('select_all_chill_nodes.sql'))).fetchall()

    node_list = result
    chill_nodes = []

    for node in node_list:
        if isinstance(node.path, str):
            chill_node = ChillNode(name=node.name, node_id=node.node_id, value=node.value, template=node.template, query=node.query, path=node.path, method=node.method, weight=node.weight)

            chill_nodes.append(chill_node)

    with open(yaml_file, 'w') as f:
        yaml.dump_all(chill_nodes, stream=f, default_flow_style=False)
        #current_app.logger.debug(yaml.dump_all(chill_nodes, default_flow_style=False))

def load_yaml(yaml_file):
    "Load ChillNode yaml objects into chill database."

    with open(yaml_file, 'r') as f:
        documents = yaml.safe_load_all(f.read())
        for item in documents:
            #current_app.logger.debug(item)
            if isinstance(item, ChillNode):
                try:
                    item.load()
                except NotImplementedError as err:
                    current_app.logger.warning(err)



# https://packaging.python.org/en/latest/distributing.html
from setuptools import setup, find_packages

__version__ = "0.8.1"  # Also set in src/chill/_version.py

setup(
    name="chill",
    version=__version__,
    author="Jake Hickenlooper",
    author_email="jake@weboftomorrow.com",
    keywords="static website SQL sqlite Flask web framework",
    description="Database driven web application framework in Flask",
    long_description="""
        This involves creating custom SQL queries to pull your data from your database
        into your jinja2 HTML templates for your website.  Chill creates a static
        version of the website or can run as a Flask app. Their are a few tables that
        are specific to Chill in order to handle page routes and what SQL query should
        be used and such.
    """,
    url="https://github.com/jkenlooper/chill",
    license="GPL",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP :: Site Management",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Software Development :: Build Tools",
    ],
    package_dir={"": "src"},
    packages=find_packages("src"),
    package_data={"chill": ["queries/*.sql"]},
    zip_safe=False,
    test_suite="chill.tests",
    install_requires=[
        "setuptools",
        "docutils",
        "Flask",
        "Jinja2",
        "Flask-Caching",
        "Frozen-Flask",
        "Flask-Markdown",
        "future",
        #"psycopg2", # Might switch to postgres in the future
        "sqlalchemy",
        "PyYAML",
        "gevent",
        "docopt",
        #'pyselect', # Using pyselect within src/chill/ instead
        "MarkupSafe",
        "babel",
    ],
    entry_points={"console_scripts": ["chill = chill.script:main"]},
)
